# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

from .validation import Validation

class Cross(Validation):
    """

    """
    def __init__(self, n_folds, stratified=False, shuffle=True):
        self.n_folds = n_folds
        self.stratified = stratified
        self.shuffle = shuffle

    def get_folds(self, data_set):
        """

        :param data_set:
        :type data_set:
        :return:
        :rtype:
        """
        return self.split_data(
            data_set,
            n_reps=1,
            stratified=self.stratified,
            shuffle=self.shuffle,
            folds_n=self.n_folds,
            fold_len=None
        )



from .parameter import Parameter
from .with_parameters import WithParameters

from .log_parameter_transformation import LogParameterTransformation
from .none_parameter_transformation import NoneParameterTransformation


# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

from .parametrizable import Parametrizable


class Parameter(Parametrizable):
    """

    """
    GRID_GROUP = 1
    OPT_GROUP = 2

    def __init__(self, name, transformation,
                 default_value, bounds=None):
        """

        :param name:
        :type name:
        :param transformation:
        :type transformation:
        :param default_value:
        :type default_value:
        :param bounds:
        :type bounds:
        """

        self.name = name
        self.transformation = transformation
        if hasattr(default_value, "__len__"):
            self.dims = len(default_value)
            self.array = self.dims > 1
            if hasattr(default_value[0], "__len__"):
                # default_value = [[1.], [1.], [1., 2.]]
                self.group = Parameter.GRID_GROUP
                self.current_value = [
                    default_value_i[0]
                    for default_value_i in default_value
                ]
                self.default_values = default_value
            else:
                # default_value = [1., 1., 1.]
                self.group = Parameter.OPT_GROUP
                self.current_value = default_value
                self.default_values = [[i] for i in default_value]
            if not self.array:
                self.current_value = self.current_value[0]
        else:
            # default_value = 1.
            self.group = Parameter.OPT_GROUP
            self.dims = 1
            self.array = False
            self.current_value = default_value
            self.default_values = [[default_value]]
        self.bounds = bounds
        if self.bounds is None:
            self.bounds = self.transformation.get_def_bounds()

    def set_param_values(self, params, only_group=None, trans=False):
        """

        :param params:
        :type params:
        :param only_group:
        :type only_group:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """
        if only_group is not None and only_group != self.group:
            return

        assert len(params) == self.dims, \
            "length of {} is not correct".format(self.name)

        if trans:
            if self.array:
                params = self.transformation.inv_trans(params).tolist()
            else:
                params = self.transformation.inv_trans(params)

        if self.array:
            self.current_value = params
        else:
            self.current_value = params[0]

    def get_param_values(self, only_group=None, trans=False):
        """

        :param only_group:
        :type only_group:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """
        if only_group is not None and only_group != self.group:
            return []

        assert self.current_value is not None, \
            "{} has not been initialized".format(self.name)

        current_value = self.current_value
        if trans:
            current_value = self.transformation.trans(current_value)
            if self.array:
                current_value = current_value.tolist()

        if self.array:
            return current_value
        return [current_value]

    def get_param_bounds(self, only_group=None, trans=False):
        """

        :param only_group:
        :type only_group:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """
        if only_group is not None and only_group != self.group:
            return []

        current_bounds = self.bounds
        if trans:
            current_bounds = tuple(
                self.transformation.trans(bound)
                for bound in current_bounds
            )

        return [current_bounds] * self.dims

    def get_param_keys(self, recursive=True, only_group=None):
        """

        :param recursive:
        :type recursive:
        :param only_group:
        :type only_group:
        :return:
        :rtype:
        """

        if only_group is not None and only_group != self.group:
            return []

        if not recursive:
            return self.name

        if self.dims == 1:
            return [self.name]

        return [
            "{}_d{}".format(self.name, dim) for dim in range(self.dims)
        ]

    def get_param_n(self, only_group=None):
        """

        :param only_group:
        :type only_group:
        :return:
        :rtype:
        """
        if only_group is not None and only_group != self.group:
            return 0

        return self.dims

    def get_default_values(self, only_group=None, trans=False):
        """

        :param only_group:
        :type only_group:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """
        if only_group is not None and only_group != self.group:
            return []

        default_values = self.default_values
        if trans:
            default_values = [
                self.transformation.trans(element).tolist()
                for element in default_values
            ]

        return default_values

    def get_group(self):
        """

        :return:
        :rtype:
        """
        return self.group

    def is_array(self):
        """

        :return:
        :rtype:
        """
        return self.array

    def grad_trans(self, df):
        """

        :param df:
        :type df:
        :return:
        :rtype:
        """

        return self.transformation.grad_trans(self.current_value, df)



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

from .covariance_function import CovarianceFunction


class Posterior(CovarianceFunction):
    """

    """
    def __init__(self, prior_gp, prior_data):
        
        self.prior_gp = prior_gp
        self.prior_data = prior_data

        super(Posterior, self).__init__(
            self.prior_gp.covariance_function.get_hyperparams(),
        )

    def covariance(self, mat_a, mat_b=None, only_diagonal=False):
        """
        Measures the distance matrix between solutions of A and B, and
        applies the kernel function element-wise to the distance matrix.

        :param mat_a: List of solutions in lines and dimensions in columns.
        :type mat_a:
        :param mat_b: List of solutions in lines and dimensions in columns.
        :type mat_b:
        :param only_diagonal:
        :type only_diagonal:
        :return: Result matrix with kernel function applied element-wise.
        :rtype:
        """

        kxa_matrix = self.prior_gp.covariance_function.marginalize_covariance(
            self.prior_data['X'],
            mat_a
        )

        if mat_b is None:
            if only_diagonal:
                kab_matrix = \
                    self.prior_gp.covariance_function.marginalize_covariance(
                        mat_a, only_diagonal=True
                    )
            else:
                kab_matrix = \
                    self.prior_gp.covariance_function.marginalize_covariance(
                        mat_a
                    )
            kxb_matrix = None
        else:
            kab_matrix = \
                self.prior_gp.covariance_function.marginalize_covariance(
                    mat_a, mat_b
                )
            kxb_matrix = \
                self.prior_gp.covariance_function.marginalize_covariance(
                    self.prior_data['X'],
                    mat_b
                )

        prior_multivariate_gaussian = self.prior_gp.marginalize_gp(
            self.prior_data['X']
        )
        covariance = prior_multivariate_gaussian.conditional_cov(
            kxa_matrix,
            kxb_matrix,
            kab_matrix
        )

        return covariance

    def dk_dx(self, mat_a, mat_b=None):
        """
        Measures gradient of the distance between solutions of A and B in X.

        :param mat_a: List of solutions in lines and dimensions in columns.
        :param mat_b: List of solutions in lines and dimensions in columns.
        :return: 3D array with the gradient in every dimension of X.
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def dk_dtheta(self, mat_a, mat_b=None, trans=False):
        """
        Measures gradient of the distance between solutions of A and B in the
        hyper-parameter space.

        :param mat_a: List of solutions in lines and dimensions in columns.
        :type mat_a:
        :param mat_b: List of solutions in lines and dimensions in columns.
        :type mat_b:
        :param trans: Return results in the transformed space.
        :type trans:
        :return: 3D array with the gradient in every
         dimension the length-scale hyper-parameter space.
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from .parameter_transformation import ParameterTransformation


class LogParameterTransformation(ParameterTransformation):
    """

    """

    @staticmethod
    def trans(x):
        """

        :param x:
        :return:
        """
        return np.log(x)

    @staticmethod
    def inv_trans(x):
        """

        :param x:
        :return:
        """
        return np.exp(x)

    @staticmethod
    def grad_trans(f, df):
        """

        :param f:
        :param df:
        :return:
        """
        return f * df

    @staticmethod
    def get_def_bounds():
        """

        :return:
        :rtype:
        """
        return np.exp(-10), np.exp(10)



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

from ..parameters.parameter import Parameter
from ..parameters.with_parameters import WithParameters


class CovarianceFunction(WithParameters):
    """

    """
    def __init__(self, hyperparams):

        super(CovarianceFunction, self).__init__(hyperparams)

    def marginalize_covariance(self, mat_a, mat_b=None,
                               only_diagonal=False,
                               dk_dx_needed=False,
                               dk_dtheta_needed=False,
                               trans=False):
        """
        Measures the distance matrix between solutions of A and B, and
        applies the kernel function element-wise to the distance matrix.

        :param mat_a: List of solutions in lines and dimensions in columns.
        :type mat_a:
        :param mat_b: List of solutions in lines and dimensions in columns.
        :type mat_b:
        :param only_diagonal:
        :type only_diagonal:
        :param dk_dx_needed: It should be true if the derivative in x is needed.
        :type dk_dx_needed:
        :param trans:
        :type trans:
        :param dk_dtheta_needed: It should be true if the derivative in
            theta is needed.
        :type dk_dtheta_needed:
        :return: Result matrix with kernel function applied element-wise.
        :rtype:
        """
        covariance = self.covariance(
            mat_a,
            mat_b=mat_b,
            only_diagonal=only_diagonal
        )

        result = (covariance, )

        if dk_dx_needed:
            dk_dx = self.dk_dx(
                mat_a,
                mat_b=mat_b
            )
            result += (dk_dx, )

        if dk_dtheta_needed:
            dk_dtheta = self.dk_dtheta(
                mat_a,
                mat_b=mat_b,
                trans=trans
            )
            result += (dk_dtheta, )

        if len(result) == 1:
            result = result[0]

        return result

    def covariance(self, mat_a, mat_b=None, only_diagonal=False):
        """
        Measures the distance matrix between solutions of A and B, and
        applies the kernel function element-wise to the distance matrix.

        :param mat_a: List of solutions in lines and dimensions in columns.
        :type mat_a:
        :param mat_b: List of solutions in lines and dimensions in columns.
        :type mat_b:
        :param only_diagonal:
        :type only_diagonal:
        :return: Result matrix with kernel function applied element-wise.
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def dk_dx(self, mat_a, mat_b=None):
        """
        Measures gradient of the distance between solutions of A and B in X.

        :param mat_a: List of solutions in lines and dimensions in columns.
        :type mat_a:
        :param mat_b: List of solutions in lines and dimensions in columns.
        :type mat_b:
        :return: 3D array with the gradient in every dimension of X.
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def dk_dtheta(self, mat_a, mat_b=None, trans=False):
        """
        Measures gradient of the distance between solutions of A and B in the
        hyper-parameter space.

        :param mat_a: List of solutions in lines and dimensions in columns.
        :type mat_a:
        :param mat_b: List of solutions in lines and dimensions in columns.
        :type mat_b:
        :param trans: Return results in the transformed space.
        :type trans:
        :return: 3D array with the gradient in every
         dimension the length-scale hyper-parameter space.
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from .covariance_function import CovarianceFunction
from ..parameters.parameter import Parameter
from ..parameters.log_parameter_transformation import LogParameterTransformation


class StationaryFunction(CovarianceFunction):
    """

    """
    def __init__(self, hyperparams, ls=None):

        default_lengthscale = ls
        if default_lengthscale is None:
            default_lengthscale = 1.

        hyperparams.append(Parameter(
            'lengthscale',
            transformation=LogParameterTransformation,
            default_value=default_lengthscale,
        ))

        super(StationaryFunction, self).__init__(hyperparams)

    def covariance(self, mat_a, mat_b=None, only_diagonal=False):
        """
        Measures the distance matrix between solutions of A and B, and
        applies the kernel function element-wise to the distance matrix.

        :param mat_a: List of solutions in lines and dimensions in columns.
        :type mat_a:
        :param mat_b: List of solutions in lines and dimensions in columns.
        :type mat_b:
        :param only_diagonal:
        :type only_diagonal:
        :return: Result matrix with kernel function applied element-wise.
        :rtype:
        """
        sq_dist = self.sq_distance(
            mat_a, mat_b=mat_b,
            only_diagonal=only_diagonal
        )

        return self.stationary_function(sq_dist)

    def dk_dx(self, mat_a, mat_b=None):
        """
        Measures gradient of the distance between solutions of A and B in X.

        :param mat_a: List of solutions in lines and dimensions in columns.
        :type mat_a:
        :param mat_b: List of solutions in lines and dimensions in columns.
        :type mat_b:
        :return: 3D array with the gradient in every dimension of X.
        :rtype:
        """
        if mat_b is None:
            mat_b = mat_a

        dr_dx = np.zeros((mat_a.shape[0], mat_b.shape[0], mat_a.shape[1]))

        for i in range(0, mat_a.shape[0]):
            dr_dx[i, :, :] = 2.0 * (mat_a[i, :] - mat_b[:, :]) / \
                              np.power(self.get_param_value('lengthscale'), 2.0)

        # TODO if covariance is needed along with dk_dx, this is computed twice
        sq_dist = self.sq_distance(mat_a, mat_b=mat_b)
        return self.dkr_dx(sq_dist, dr_dx)

    def dk_dtheta(self, mat_a, mat_b=None, trans=False):
        """
        Measures gradient of the distance between solutions of A and B in the
        hyper-parameter space.

        :param mat_a: List of solutions in lines and dimensions in columns.
        :type mat_a:
        :param mat_b: List of solutions in lines and dimensions in columns.
        :type mat_b:
        :param trans: Return results in the transformed space.
        :type trans:
        :return: 3D array with the gradient in every
         dimension the length-scale hyper-parameter space.
        :rtype:
        """
        if mat_b is None:
            mat_b = mat_a

        # TODO if covariance is needed along with dk_dx, this is computed twice
        sq_dist = self.sq_distance(mat_a, mat_b=mat_b)
        result = self.dkr_dtheta(sq_dist, trans)

        if not self.get_hyperparam('lengthscale').\
                get_group() == Parameter.OPT_GROUP:
            return result

        dr_dl = np.zeros((mat_a.shape[0], mat_b.shape[0], mat_a.shape[1]))
        for i in range(0, mat_a.shape[0]):
            dr_dl[i, :, :] = - 2.0 * (
                np.power(mat_a[i, :] - mat_b[:, :], 2.0)) / \
                    np.power(self.get_param_value('lengthscale'), 3.0)

        dkr_dl = self.dkr_dl(sq_dist, dr_dl)

        if trans:
            dkr_dl = self.get_hyperparam('lengthscale').grad_trans(
                dkr_dl
            )

        l_len = self.get_hyperparam('lengthscale').get_param_n()

        if l_len == 1:
            result += np.sum(dkr_dl, axis=2),
        elif l_len > 1:
            for i in range(l_len):
                result += dkr_dl[:, :, i],
        else:
            raise Exception("lengthscale array length not valid")

        return result

    def sq_distance(self, mat_a, mat_b=None, only_diagonal=False):
        """
        Measures the distance matrix between solutions of A and B.

        :param mat_a: List of solutions in lines and dimensions in columns.
        :type mat_a:
        :param mat_b: List of solutions in lines and dimensions in columns.
        :type mat_b:
        :param only_diagonal:
        :type only_diagonal:
        :return: Distance matrix between solutions of A and B.
        :rtype:
        """
        lengthscale = self.get_hyperparam('lengthscale')
        param_values = lengthscale.get_param_values()

        if not lengthscale.is_array():
            param_values = param_values[0]

        mat_a = mat_a / param_values
        if mat_b is None:
            if only_diagonal:
                return np.zeros((len(mat_a), 1))
            center = np.mean(mat_a, axis=0)
            mat_a = mat_a - center
            mat_a_sq = np.sum(np.square(mat_a), 1)

            r2 = mat_a_sq[:, None] + mat_a_sq[None, :]

            r2 -= 2. * np.dot(mat_a, mat_a.T)

            np.lib.stride_tricks.as_strided(
                r2, shape=(r2.shape[0], ),
                strides=((r2.shape[0] + 1) * r2.itemsize, )
            )[:, ] = 0
        else:
            mat_b = mat_b / param_values
            center = np.mean(np.vstack((mat_a, mat_b)), axis=0)
            mat_a = mat_a - center
            mat_b = mat_b - center
            mat_a_sq = np.sum(np.square(mat_a), 1)
            mat_b_sq = np.sum(np.square(mat_b), 1)

            r2 = mat_a_sq[:, None] + mat_b_sq[None, :]

            r2 -= 2. * np.dot(mat_a, mat_b.T)
        r2 = r2.clip(min=0.0)

        return r2

    def stationary_function(self, sq_dist):
        """
        It applies the kernel function
        element-wise to the distance matrix.

        :param sq_dist: Distance matrix
        :type sq_dist:
        :return: Result matrix with kernel function applied element-wise.
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def dkr_dx(self, sq_dist, dr_dx):
        """
        Measures gradient of the kernel function in X.

        :param sq_dist: Square distance
        :type sq_dist:
        :param dr_dx:
        :type dr_dx:
        :return: 3D array with the gradient of the kernel function in every
         dimension of X.
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def dkr_dtheta(self, sq_dist, trans):
        """
        Measures gradient of the kernel function in the
        hyper-parameter space.

        :param sq_dist: Square distance
        :type sq_dist:
        :param trans: Return results in the transformed space.
        :type trans:
        :return: 3D array with the gradient of the kernel function in every
         dimension the length-scale hyper-parameter space.
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def dkr_dl(self, sq_dist, dr_dl):
        """
        Measures gradient of the kernel function in the
        hyper-parameter space.

        :param sq_dist: Square distance
        :type sq_dist:
        :param dr_dl:
        :type dr_dl:
        :return: 3D array with the gradient of the kernel function in every
         dimension the length-scale hyper-parameter space.
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")



# -*- coding: utf-8 -*-
#
#    Copyright 2020 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import scipy.linalg


class MultivariateGaussian(object):
    """
    Multivariate Gaussian distribution
    """
    def __init__(self, mean, covariance,
                 cached_cov_matrix=None, cached_l_matrix=None):
        self.mean = mean
        self.covariance = covariance
        self.l_matrix = MultivariateGaussian._get_l_matrix(
            covariance,
            cached_cov_matrix,
            cached_l_matrix
        )

    def sample(self, n_samples=10):
        """

        :param n_samples:
        :type n_samples:
        :return:
        :rtype:
        """
        rnd = np.random.randn(self.l_matrix.shape[0], n_samples)

        return self.mean.reshape(-1, 1) + np.dot(self.l_matrix, rnd)

    def conditional_mean(self, data_y, ka_matrix, test_mean):
        """

        :param data_y:
        :type data_y:
        :param ka_matrix:
        :type ka_matrix:
        :param test_mean:
        :type test_mean:
        :return:
        :rtype:
        """
        alpha = scipy.linalg.cho_solve(
            (self.l_matrix, True),
            data_y - self.mean
        )

        new_mean = np.dot(ka_matrix.T, alpha) + test_mean

        return new_mean

    def conditional_cov(self, kxa_matrix, kxb_matrix, kab_matrix):
        """

        :param kxa_matrix:
        :type kxa_matrix:
        :param kxb_matrix:
        :type kxb_matrix:
        :param kab_matrix:
        :type kab_matrix:
        :return:
        :rtype:
        """

        beta_a = scipy.linalg.solve_triangular(
            self.l_matrix,
            kxa_matrix,
            lower=True,
            check_finite=False
        )

        if kxb_matrix is not None:
            beta_b = scipy.linalg.solve_triangular(
                self.l_matrix,
                kxb_matrix,
                lower=True,
                check_finite=False
            )
        else:
            beta_b = beta_a

        if kab_matrix.shape[0] == kab_matrix.shape[1]:
            new_covariance = kab_matrix - np.dot(beta_a.T, beta_b)
        else:
            new_covariance = kab_matrix - \
                np.sum(np.power(beta_a, 2), axis=0).reshape(-1, 1)
            new_covariance = new_covariance.clip(min=1e-20)

        return new_covariance

    def get_log_likelihood(self, data_y, dmu_dtheta=None, dk_dtheta=None):
        """

        :param data_y:
        :type data_y:
        :param dmu_dtheta:
        :type dmu_dtheta:
        :param dk_dtheta:
        :type dk_dtheta:
        :return:
        :rtype:
        """
        alpha = scipy.linalg.cho_solve(
            (self.l_matrix, True),
            data_y - self.mean
        )
        
        log_likelihood = self._log_likelihood(data_y, alpha)

        result = (log_likelihood, )

        if dmu_dtheta is not None and dk_dtheta is not None:
            dlog_likelihood_dtheta = self._dlog_likelihood_dtheta(
                alpha,
                dmu_dtheta,
                dk_dtheta
            )
            result += (dlog_likelihood_dtheta, )

        if len(result) == 1:
            result = result[0]

        return result

    def get_log_predictive_density(self, data_y):
        """
        Measure the log predictive density

        :param data_y:
        :type data_y:
        :return:
        :rtype:
        """
        var = np.diagonal(self.covariance)[:, None]
        return MultivariateGaussian._log_density(data_y, var, self.mean)

    def get_log_loocv_density(self, data_y):
        """
        Measure the log predictive density

        :param data_y:
        :type data_y:
        :return:
        :rtype:
        """
        inv_cov = scipy.linalg.cho_solve(
            (self.l_matrix, True),
            np.eye(self.l_matrix.shape[0])
        )

        diag_inv_cov = np.diagonal(inv_cov)[:, None]

        mean = data_y - np.divide(np.dot(inv_cov, data_y), diag_inv_cov)

        var = np.divide(1.0, diag_inv_cov)

        return MultivariateGaussian._log_density(data_y, var, mean)

    def _log_likelihood(self, data_y, alpha):
        """
        Measure the log Likelihood

        :param data_y:
        :type data_y:
        :param alpha:
        :type alpha:
        :return:
        :rtype:
        """
        llikelihood = -np.sum(np.log(np.diag(self.l_matrix))) - \
            0.5 * np.dot((data_y - self.mean).T, alpha)[0, 0] - \
            0.5 * self.l_matrix.shape[0] * np.log(2.0 * np.pi)

        return llikelihood

    def _dlog_likelihood_dtheta(self, alpha, dmu_dtheta, dk_dtheta):
        """
        Measure the gradient log Likelihood

        :param alpha:
        :type alpha:
        :param dmu_dtheta:
        :type dmu_dtheta:
        :param dk_dtheta:
        :type dk_dtheta:
        :return:
        :rtype:
        """
        k_inv = scipy.linalg.cho_solve(
            (self.l_matrix, True),
            np.eye(self.l_matrix.shape[0]))
        jacobian = np.outer(alpha, alpha) - k_inv

        grad_llikelihood = []
        # Log amplitude gradient.
        for dmu_dtheta_i in dmu_dtheta:
            grad_llikelihood.append(
                0.5 * np.trace(np.dot(jacobian, dmu_dtheta_i))
            )
        # Log amplitude gradient.
        for dk_dtheta_i in dk_dtheta:
            grad_llikelihood.append(
                0.5 * np.trace(np.dot(jacobian, dk_dtheta_i))
            )

        return grad_llikelihood

    @staticmethod
    def _log_density(fy, var, mean):
        """
        Measure the log density

        :param fy:
        :type fy:
        :param var:
        :type var:
        :param mean:
        :type mean:
        :return:
        :rtype:
        """
        log_predictive_density = \
            -0.5 * np.log(var) - \
            np.power((fy - mean), 2) / (2.0 * var) - \
            0.5 * np.log(2 * np.pi)

        return log_predictive_density

    @staticmethod
    def _get_l_matrix(covariance, cached_cov_matrix, cached_l_matrix=None):
        """

        :param covariance:
        :type covariance:
        :param cached_cov_matrix:
        :type cached_cov_matrix:
        :param cached_l_matrix:
        :type cached_l_matrix:
        :return:
        :rtype:
        """
        if cached_l_matrix is None:
            return MultivariateGaussian._chol(covariance)

        if covariance is None:
            return cached_l_matrix

        if cached_l_matrix.shape[0] < covariance.shape[0]:
            return MultivariateGaussian._update_chol(
                covariance,
                cached_l_matrix,
                offset=0
            )

        return MultivariateGaussian._downdate_chol(
            cached_cov_matrix,
            cached_l_matrix,
            offset=0,
            size=(cached_l_matrix.shape[0] - covariance.shape[0])
        )

    @staticmethod
    def _chol(k_matrix):
        """
        Compute cholesky decomposition

        :param k_matrix:
        :type k_matrix:
        :return:
        :rtype:
        """
        MultivariateGaussian._check_k_matrix(k_matrix)

        # Solve cholesky decomposition
        l_matrix = MultivariateGaussian._jittered_chol(k_matrix)

        MultivariateGaussian._check_l_matrix(l_matrix, k_matrix)

        return l_matrix

    @staticmethod
    def _update_chol(k_matrix, l_matrix, offset=0):
        """

        :param k_matrix:
        :type k_matrix:
        :param l_matrix:
        :type l_matrix:
        :param offset:
        :type offset:
        :return:
        :rtype:
        """
        MultivariateGaussian._check_k_matrix(k_matrix)

        b_i = k_matrix.shape[0] - offset
        new_i = l_matrix.shape[0] - offset
        new_l_matrix = np.zeros(k_matrix.shape)
        new_l_matrix[:new_i, :new_i] = l_matrix[:new_i, :new_i]
        new_l_matrix[new_i:b_i:, :new_i] = scipy.linalg.solve_triangular(
            l_matrix[:new_i, :new_i],
            k_matrix[:new_i, new_i:b_i],
            lower=True,
            check_finite=False
        ).T

        new_l_matrix[b_i:, :new_i] = l_matrix[new_i:, :new_i]
        new_l_matrix[new_i:b_i, new_i:b_i] = \
            MultivariateGaussian._jittered_chol(
                k_matrix[new_i:b_i, new_i:b_i] -
                np.dot(
                    new_l_matrix[new_i:b_i:, :new_i],
                    new_l_matrix[new_i:b_i:, :new_i].T
                )
            )
        new_l_matrix[b_i:, new_i:b_i] = scipy.linalg.solve_triangular(
            new_l_matrix[new_i:b_i, new_i:b_i],
            k_matrix[b_i:, new_i:b_i].T -
            np.dot(
                new_l_matrix[new_i:b_i:, :new_i],
                new_l_matrix[b_i:, :new_i].T
            ),
            lower=True,
            check_finite=False
        ).T
        new_l_matrix[b_i:, b_i:] = MultivariateGaussian._jittered_chol(
            np.dot(l_matrix[new_i:, new_i:], l_matrix[new_i:, new_i:].T) -
            np.dot(
                new_l_matrix[b_i:, new_i:b_i],
                new_l_matrix[b_i:, new_i:b_i].T
            )
        )

        MultivariateGaussian._check_l_matrix(new_l_matrix, k_matrix)

        return new_l_matrix

    @staticmethod
    def _downdate_chol(k_matrix, l_matrix, offset=0, size=0):
        """

        :param k_matrix:
        :type k_matrix:
        :param l_matrix:
        :type l_matrix:
        :param offset:
        :type offset:
        :param size:
        :type size:
        :return:
        :rtype:
        """
        MultivariateGaussian._check_k_matrix(k_matrix)

        b_i = k_matrix.shape[0] - offset
        new_i = b_i - size
        new_size = k_matrix.shape[0] - size
        new_l_matrix = np.zeros((new_size, new_size))
        new_l_matrix[:new_i, :new_i] = l_matrix[:new_i, :new_i]
        new_l_matrix[new_i:, new_i:] = MultivariateGaussian._jittered_chol(
            np.dot(l_matrix[b_i:, b_i:], l_matrix[b_i:, b_i:].T) +
            np.dot(l_matrix[b_i:, new_i:b_i], l_matrix[b_i:, new_i:b_i].T)
        )
        new_l_matrix[new_i:, :new_i] = l_matrix[b_i:, :new_i]

        new_k_matrix = np.vstack((
            np.hstack((k_matrix[:new_i, :new_i], k_matrix[:new_i, b_i:])),
            np.hstack((k_matrix[b_i:, :new_i], k_matrix[b_i:, b_i:]))
        ))

        MultivariateGaussian._check_l_matrix(new_l_matrix, new_k_matrix)

        return new_l_matrix

    @staticmethod
    def _check_k_matrix(k_matrix):
        """

        :param k_matrix:
        :type k_matrix:
        :return:
        :rtype:
        """

        # Empty
        if k_matrix.size == 0:
            return np.empty((0, 0))

        # Non finite values in covariance matrix
        if not np.all(np.isfinite(k_matrix)):
            raise np.linalg.LinAlgError("Non finite values in cov matrix")

        # Covariance matrix is not symmetric
        if np.max(np.abs(k_matrix - k_matrix.T)) > 1e-20:
            raise np.linalg.LinAlgError("Covariance matrix is not symmetric")

    @staticmethod
    def _jittered_chol(k_matrix):
        """

        :param k_matrix:
        :type k_matrix:
        :return:
        :rtype:
        """

        l_matrix = None
        jitter = 1e-30
        max_jitter = 1e-6
        k_corrected = k_matrix
        while l_matrix is None and jitter < max_jitter:
            l_matrix, error = scipy.linalg.lapack.dpotrf(
                np.ascontiguousarray(k_corrected),
                lower=1
            )
            if error != 0:
                l_matrix = None
                k_corrected = k_matrix + jitter * np.eye(k_matrix.shape[0])
                jitter *= 10

        if l_matrix is None:
            raise np.linalg.LinAlgError("Can't compute cholesky decomposition")

        return l_matrix

    @staticmethod
    def _check_l_matrix(l_matrix, k_matrix):
        """

        :param l_matrix:
        :type l_matrix:
        :param k_matrix:
        :type k_matrix:
        :return:
        :rtype:
        """

        # Non finite values in L matrix
        if not np.all(np.isfinite(l_matrix)):
            raise np.linalg.LinAlgError("Non finite values in L matrix")

        # Main diagonal of L not positive
        if np.min(np.diagonal(l_matrix)) <= 0.0:
            raise np.linalg.LinAlgError("Main diagonal of L not positive")

        # Errors in L matrix multiplication
        if np.max(np.abs(k_matrix - np.dot(l_matrix, l_matrix.T))) > 1e-4:
            raise np.linalg.LinAlgError("Errors in L matrix multiplication")



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from .parameters.with_parameters import WithParameters
from .covariance_functions.posterior import Posterior as PosteriorCov
from .covariance_functions.white_noise import WhiteNoise
from .mean_functions.posterior import Posterior as PosteriorMean
from .multivariate_gaussian import MultivariateGaussian


class GaussianProcess(WithParameters):
    """
    GPlib module for Gaussian Process
    """
    def __init__(self, mean_function, covariance_function,
                 noise=True, cache_size=1):
        self.mean_function = mean_function
        self.covariance_function = covariance_function
        self.cache = {
            'size': cache_size,
            'gp_hashes': [],
            'data': [],
            'mgds': []
        }
        hyperparameters = [mean_function, covariance_function]
        if noise:
            self.white_noise = WhiteNoise(ov2=1e-3)
            hyperparameters.append(self.white_noise)
        else:
            self.white_noise = None

        super(GaussianProcess, self).__init__(hyperparameters)

    def get_posterior(self, data):
        """

        :param data:
        :type data:
        :return:
        :rtype:
        """
        return GaussianProcess(
            PosteriorMean(self, data),
            PosteriorCov(self, data)
        )

    def sample(self, test_points, n_samples=10):
        """
        Sample the prior of the GP

        :param test_points:
        :type test_points:
        :param n_samples:
        :type n_samples:
        :return:
        :rtype:
        """
        multivariate_gaussian = self.marginalize_gp(test_points)
        return multivariate_gaussian.sample(n_samples)

    def get_log_likelihood(self, data, grad_needed=False):
        """

        :param data:
        :type data:
        :param grad_needed:
        :type grad_needed:
        :return:
        :rtype:
        """

        multivariate_gaussian = self.marginalize_gp(data['X'])
        dmu_dtheta = None
        dk_dtheta = None
        if grad_needed:
            _, dmu_dtheta = self.mean_function.marginalize_mean(
                data['X'], dmu_dtheta_needed=True, trans=True)
            _, dk_dtheta = self.covariance_function.marginalize_covariance(
                data['X'], dk_dtheta_needed=True, trans=True)
            if self.white_noise is not None:
                _, dn_dtheta = self.white_noise.marginalize_covariance(
                    data['X'], dk_dtheta_needed=True, trans=True)
                dk_dtheta += dn_dtheta
        return multivariate_gaussian.get_log_likelihood(
            data['Y'],
            dmu_dtheta,
            dk_dtheta
        )

    def get_log_predictive_density(self, data):
        """
        Measure the log predictive density

        :param data:
        :type data:
        :return:
        :rtype:
        """
        multivariate_gaussian = self.marginalize_gp(data['X'])
        return multivariate_gaussian.get_log_predictive_density(
            data['Y']
        )

    def get_log_loocv_density(self, data):
        """
        Measure the log predictive density

        :param data:
        :type data:
        :return:
        :rtype:
        """
        multivariate_gaussian = self.marginalize_gp(data['X'])
        return multivariate_gaussian.get_log_loocv_density(
            data['Y']
        )

    def marginalize_gp(self, data_x):
        """
        Get mean and covariance of following points

        :param data_x:
        :type data_x:
        :return:
        :rtype:
        """

        # Data assertions
        assert not np.isnan(data_x).any(), "NaN values in X"
        assert not np.isinf(data_x).any(), "Inf values in X"

        gp_hash = hash(self)

        best_i = None
        best_size_diff = np.inf
        for cache_i in range(len(self.cache['gp_hashes'])):
            cached_gp_hash = self.cache['gp_hashes'][cache_i]
            if cached_gp_hash == gp_hash:
                cached_data = self.cache['data'][cache_i]

                data_size = min(cached_data.shape[0], data_x.shape[0])
                cached_data_hash = hash(cached_data[:data_size, :].tostring())
                data_x_hash = hash(data_x[:data_size, :].tostring())

                size_diff = np.abs(cached_data.shape[0] - data_x.shape[0])

                if cached_data_hash == data_x_hash and \
                        size_diff < best_size_diff:
                    best_i = cache_i
                    best_size_diff = size_diff

        if best_i is not None:
            cached_data = self.cache['data'][best_i]
            cached_l_mgd = self.cache['mgds'][best_i]
            cached_l_matrix = cached_l_mgd.l_matrix
            if cached_data.shape[0] == data_x.shape[0]:
                return cached_l_mgd
            elif cached_data.shape[0] < data_x.shape[0]:
                cached_cov_matrix = None
            else:
                cached_cov_matrix = \
                    self.covariance_function.marginalize_covariance(
                        cached_data
                    )
        else:
            cached_l_matrix = None
            cached_cov_matrix = None
        covariance = self.covariance_function.marginalize_covariance(data_x)
        if self.white_noise is not None:
            covariance += self.white_noise.marginalize_covariance(data_x)
        mean = self.mean_function.marginalize_mean(data_x)
        mgd = MultivariateGaussian(
            mean,
            covariance,
            cached_cov_matrix,
            cached_l_matrix
        )

        self.cache['gp_hashes'].append(gp_hash)
        self.cache['data'].append(data_x)
        self.cache['mgds'].append(mgd)

        if len(self.cache['gp_hashes']) > self.cache['size']:
            del self.cache['gp_hashes'][0]
            del self.cache['data'][0]
            del self.cache['mgds'][0]

        return mgd
